@@ -6,108 +6,20 @@
 #include "pch.h"
 #include "engine.h"
 
+#include "textedit.h"
+
 static bool layoutpass, actionon = false;
 static int mousebuttons = 0;
-static g3d_gui *windowhit = NULL;
+static struct gui *windowhit = NULL;
 
-//text field state - only one ever active/focused
-static string fieldname, fieldtext;  //copy of while focused
-static int fieldpos = -1, fieldlen = 0; //-1=no focus, -2=wanting to commit
-static bool fieldactive; 
+static float firstx, firsty;
 
-bool menukey(int code, bool isdown, int cooked) 
-{
-    if(code==-1 && g3d_windowhit(isdown, true)) return true;  
-    else if(code==-3 && g3d_windowhit(isdown, false)) return true;
-    
-	if(fieldpos<0) return false;
-	switch(code) 
-	{
-		case SDLK_RETURN:
-        case SDLK_KP_ENTER:
-			fieldpos = -2; //signal field commit
-			return false;
-        case SDLK_HOME:
-        case SDLK_END:
-		case SDLK_DELETE:
-        case SDLK_BACKSPACE:
-        case SDLK_LEFT:
-        case SDLK_RIGHT:
-		case SDLK_TAB:
-			break;
-		default:
-			if(!cooked || (code<32) || (code>127)) return false;
-	}
-	if(!isdown) return true;	
-	int len = strlen(fieldtext);
-    if(fieldpos>len) fieldpos = len;
-	switch(code) 
-    {
-        case SDLK_HOME:
-            fieldpos = 0;
-            break;
-        case SDLK_END:
-            fieldpos = len;
-            break;
-		case SDLK_LEFT:
-			if(fieldpos > 0) fieldpos--;
-			break;
-		case SDLK_RIGHT:
-		case SDLK_TAB:
-			if(fieldpos < len) fieldpos++;
-			break;
-		case SDLK_DELETE:
-            if(fieldpos < len) 
-            {
-                memmove(fieldtext+fieldpos, fieldtext+fieldpos+1, len-fieldpos);
-                fieldtext[len-1] = '\0';
-            }
-			break;
-		case SDLK_BACKSPACE:
-            if(fieldpos > 0) 
-            {   
-                fieldpos--;
-                memmove(fieldtext+fieldpos, fieldtext+fieldpos+1, len-fieldpos);
-                fieldtext[len-1] = '\0';
-            }
-			break;
-		default:
-            if(fieldpos<fieldlen)
-            {
-                memmove(fieldtext+fieldpos+1, fieldtext+fieldpos, len-fieldpos); //length then limited inside field draw code
-			    fieldtext[fieldpos++] = cooked;
-                fieldtext[len+1] = '\0';
-            }
-            break;
-	}
-	return true;
-}
-
-VARP(gui2d, 0, 0, 1);
+static enum {FIELDCOMMIT, FIELDABORT, FIELDEDIT, FIELDSHOW} fieldmode = FIELDSHOW; 
+static bool fieldsactive = false;
 
 static bool hascursor;
 static float cursorx = 0.5f, cursory = 0.5f;
 
-void g3d_cursorpos(float &x, float &y)
-{
-    if(gui2d) { x = cursorx; y = cursory; }
-    else x = y = 0.5f;
-}
-
-void g3d_resetcursor()
-{
-    cursorx = cursory = 0.5f;
-}
-
-bool g3d_movecursor(int dx, int dy)
-{
-    if(!gui2d || !hascursor) return false;
-    const float CURSORSCALE = 500.0f;
-    cursorx = max(0, min(1, cursorx+dx/CURSORSCALE));
-    cursory = max(0, min(1, cursory+dy/CURSORSCALE));
-    return true;
-}
-
 #define SHADOW 4
 #define ICON_SIZE (FONTH-SHADOW)
 #define SKIN_W 256
@@ -116,6 +28,7 @@ bool g3d_movecursor(int dx, int dy)
 #define INSERT (3*SKIN_SCALE)
 
 VARP(guiautotab, 6, 16, 40);
+VARP(guiclicktab, 0, 0, 1);
 
 struct gui : g3d_gui
 {
@@ -129,6 +42,7 @@ struct gui : g3d_gui
     static vector<list> lists;
     static float hitx, hity;
     static int curdepth, curlist, xsize, ysize, curx, cury;
+    static bool shouldmergehits, shouldautotab;
 
     static void reset()
     {
@@ -137,22 +51,58 @@ struct gui : g3d_gui
 
     static int ty, tx, tpos, *tcurrent, tcolor; //tracking tab size and position since uses different layout method...
 
+    void allowautotab(bool on)
+    {
+        shouldautotab = on;
+    }
+
     void autotab() 
     { 
         if(tcurrent)
         {
             if(layoutpass && !tpos) tcurrent = NULL; //disable tabs because you didn't start with one
-            if(!curdepth && (layoutpass ? 0 : cury) + ysize > guiautotab*FONTH) tab(NULL, tcolor); 
+            if(shouldautotab && !curdepth && (layoutpass ? 0 : cury) + ysize > guiautotab*FONTH) tab(NULL, tcolor); 
         }
     }
 
+    bool shouldtab()
+    {
+        if(tcurrent && shouldautotab)
+        {
+            if(layoutpass)
+            {
+                int space = guiautotab*FONTH - ysize;
+                if(space < 0) return true;
+                int l = lists[curlist].parent;
+                while(l >= 0)
+                {
+                    space -= lists[l].h;
+                    if(space < 0) return true;
+                    l = lists[l].parent;
+                }
+            }
+            else
+            {
+                int space = guiautotab*FONTH - cury;
+                if(ysize > space) return true;
+                int l = lists[curlist].parent;
+                while(l >= 0)
+                {
+                    if(lists[l].h > space) return true;
+                    l = lists[l].parent;
+                }
+            }
+        }
+        return false;
+    }
+
     bool visible() { return (!tcurrent || tpos==*tcurrent) && !layoutpass; }
 
     //tab is always at top of page
     void tab(const char *name, int color) 
     {
         if(curdepth != 0) return;
-        tcolor = color;
+        if(color) tcolor = color;
         tpos++; 
         if(!name) 
         {
@@ -175,14 +125,14 @@ struct gui : g3d_gui
                 y1 = cury - ((skiny[5]-skiny[1])-(skiny[3]-skiny[2]))*SKIN_SCALE-h,
                 y2 = cury;
             bool hit = tcurrent && windowhit==this && hitx>=x1 && hity>=y1 && hitx<x2 && hity<y2;
-            if(hit) 
+            if(hit && (!guiclicktab || mousebuttons&G3D_DOWN)) 
             {	
                 *tcurrent = tpos; //roll-over to switch tab
                 color = 0xFF0000;
             }
             
             skin_(x1-skinx[visible()?2:6]*SKIN_SCALE, y1-skiny[1]*SKIN_SCALE, w, h, visible()?10:19, 9);
-            text_(name, x1 + (skinx[3]-skinx[2])*SKIN_SCALE - INSERT, y1 + (skiny[2]-skiny[1])*SKIN_SCALE - INSERT, color, visible());
+            text_(name, x1 + (skinx[3]-skinx[2])*SKIN_SCALE - INSERT, y1 + (skiny[2]-skiny[1])*SKIN_SCALE - INSERT, tcolor, visible());
         }
         tx += w + ((skinx[5]-skinx[4]) + (skinx[3]-skinx[2]))*SKIN_SCALE; 
     }
@@ -270,8 +220,11 @@ struct gui : g3d_gui
         }
     }
 
+    void mergehits(bool on) { shouldmergehits = on; }
+
     bool ishit(int w, int h, int x = curx, int y = cury)
     {
+        if(shouldmergehits) return windowhit==this && (ishorizontal() ? hitx>=x && hitx<x+w : hity>=y && hity<y+h);
         if(ishorizontal()) h = ysize;
         else w = xsize;
         return windowhit==this && hitx>=x && hity>=y && hitx<x+w && hity<y+h;
@@ -286,12 +239,12 @@ struct gui : g3d_gui
         return layout(size+SHADOW, size+SHADOW);
     }
     
-    int texture(Texture *t, float scale)
+    int texture(Texture *t, float scale, int rotate, int xoff, int yoff, Texture *glowtex, const vec &glowcolor)
     {
         autotab();
         if(scale==0) scale = 1;
         int size = (int)(scale*2*FONTH)-SHADOW;
-        if(t!=notexture && visible()) icon_(t, true, true, curx, cury, size, ishit(size+SHADOW, size+SHADOW));
+        if(t!=notexture && visible()) icon_(t, true, true, curx, cury, size, ishit(size+SHADOW, size+SHADOW), rotate, xoff, yoff, glowtex, glowcolor);
         return layout(size+SHADOW, size+SHADOW);
     }
 
@@ -317,12 +270,12 @@ struct gui : g3d_gui
             {
                 hit = ishit(FONTH, ysize, x, y);
                 px = x + (FONTH-w)/2;
-                py = y + (ysize-FONTH) - ((ysize-FONTH)*(val-vmin))/max(vmax-vmin, 1); //zero at the bottom
+                py = y + (ysize-FONTH) - ((ysize-FONTH)*(val-vmin))/((vmax==vmin) ? 1 : (vmax-vmin)); //vmin at bottom
             }
             else
             {
                 hit = ishit(xsize, FONTH, x, y);
-                px = x + ((xsize-w)*(val-vmin))/max(vmax-vmin, 1);
+                px = x + FONTH/2 - w/2 + ((xsize-w)*(val-vmin))/((vmax==vmin) ? 1 : (vmax-vmin)); //vmin at left
                 py = y;
             }
 
@@ -330,96 +283,105 @@ struct gui : g3d_gui
             text_(label, px, py, color, hit && actionon);
             if(hit && actionon)
             {
-                int vnew = 1+vmax-vmin;
-                if(ishorizontal()) vnew = int(vnew*(y+ysize-hity)/ysize);
-                else vnew = int(vnew*(hitx-x)/xsize);
+                int vnew = (vmin < vmax ? 1 : -1)+vmax-vmin;
+                if(ishorizontal()) vnew = int(vnew*(y+ysize-FONTH/2-hity)/(ysize-FONTH));
+                else vnew = int(vnew*(hitx-x-FONTH/2)/(xsize-w));
                 vnew += vmin;
+                vnew = vmin < vmax ? clamp(vnew, vmin, vmax) : clamp(vnew, vmax, vmin);
                 if(vnew != val) val = vnew;
             }
         }
     }
 
-    char *field(const char *name, int color, int length, const char *initval)
-	{	
-        length = min(length, (int)sizeof(string)-1);
-        int w = char_width('%')*length + FONTW;
+    char *field(const char *name, int color, int length, int height, const char *initval, int initmode)
+    {	
+        editor *e = useeditor(name, initmode, false, initval); // generate a new editor if necessary
+        if(layoutpass)
+        {
+            if(initval && e->mode==EDITORFOCUSED && (e!=currentfocus() || fieldmode == FIELDSHOW))
+            {
+                if(strcmp(e->lines[0].text, initval)) e->clear(initval);
+            }
+            e->linewrap = (length<0);
+            e->maxx = (e->linewrap) ? -1 : length;
+            e->maxy = (height<=0)?1:-1;
+            e->pixelwidth = abs(length)*FONTW;
+            if(e->linewrap && e->maxy==1) 
+            {
+                int temp;
+                text_bounds(e->lines[0].text, temp, e->pixelheight, e->pixelwidth); //only single line editors can have variable height
+            }
+            else 
+                e->pixelheight = FONTH*max(height, 1); 
+        }
+        int h = e->pixelheight;
+        int w = e->pixelwidth + FONTW;
+        
+        bool wasvertical = isvertical();
+        if(wasvertical && e->maxy != 1) pushlist();
+        
         char *result = NULL;
         if(visible() && !layoutpass)
-		{
-            bool hit = ishit(w, FONTH), editing = !strcmp(fieldname, name);            
-            if(hit && (mousebuttons&G3D_DOWN) && !editing) //mouse request focus
+        {
+            bool hit = ishit(w, h);
+            if(hit) 
             {
-                s_strcpy(fieldname, name);
-                s_strcpy(fieldtext, initval);
-                fieldtext[length] = '\0';
-                fieldpos = strlen(fieldtext);
-                fieldlen = length;
-                editing = true;
-            }
-            if(editing && windowhit==this)
-            {
-                if(fieldpos==-2) 
-                {
-                    result = fieldtext;
-                    fieldpos = -1;
-                    fieldname[0] = '\0';
-                    editing = false;
+                if(mousebuttons&G3D_DOWN) //mouse request focus
+                {   
+                    useeditor(name, initmode, true); 
+                    e->mark(false);
+                    fieldmode = FIELDEDIT;
                 } 
-                else if(hit) fieldactive = true;
-                else fieldpos = -1; // mouse wandered out of focus
             }
-            if(editing && hit && (mousebuttons&G3D_PRESSED)) //mouse request position
+            bool editing = (fieldmode != FIELDSHOW) && (e==currentfocus());
+            if(hit && editing && (mousebuttons&G3D_PRESSED)!=0) e->hit(int(floor(hitx-(curx+FONTW/2))), int(floor(hity-cury)), (mousebuttons&G3D_DRAGGED)!=0); //mouse request position
+            if(editing && ((fieldmode==FIELDCOMMIT) || (fieldmode==FIELDABORT) || !hit)) // commit field if user pressed enter or wandered out of focus 
             {
-                int x = curx+FONTW/2;
-                fieldpos = 0;
-                while(fieldtext[fieldpos] && fieldpos < length)
-                {
-                    int nx = char_width(fieldtext[fieldpos], x);
-                    if(nx > hitx) break;
-                    x = nx;
-                    fieldpos++;
-                }
-            }
-                           
+                if(fieldmode==FIELDCOMMIT || (fieldmode!=FIELDABORT && !hit)) result = e->currentline().text;
+                e->active = (e->mode!=EDITORFOCUSED);
+                fieldmode = FIELDSHOW;
+            } 
+            else fieldsactive = true;
+            
+            e->draw(curx+FONTW/2, cury, color, hit && editing);
+            
             notextureshader->set();
             glDisable(GL_TEXTURE_2D);
-			if(editing) glColor3f(1, 0, 0);
+            if(editing) glColor3f(1, 0, 0);
             else glColor3ub(color>>16, (color>>8)&0xFF, color&0xFF);
-			glBegin(GL_LINE_LOOP);
-            rect_(curx, cury, w, FONTH);
-			glEnd();
+            glBegin(GL_LINE_LOOP);
+            rect_(curx, cury, w, h);
+            glEnd();
             glEnable(GL_TEXTURE_2D);
             defaultshader->set();
-            
-            draw_text(editing ? fieldtext : (result ? result : initval), curx+FONTW/2, cury, color>>16, (color>>8)&0xFF, color&0xFF);
-            
-            if(editing && hit && fieldpos>=0 && (totalmillis/250)&1) 
-            {
-                int fx = curx+FONTW/2 + text_width(fieldtext, fieldpos);
-                glColor3f(1, 0, 0);
-                notextureshader->set();
-                glDisable(GL_TEXTURE_2D);
-                glBegin(GL_QUADS);
-                rect_(fx-2, cury, 4, FONTH);
-                glEnd();				
-                glEnable(GL_TEXTURE_2D);
-                defaultshader->set();
-            }
         }
-    	layout(w, FONTH);
-		return result;
-	}
+        layout(w, h);
+        
+        if(e->maxy != 1)
+        {
+            int slines = e->lines.length()-e->pixelheight/FONTH;
+            if(slines > 0) 
+            {
+                int pos = e->scrolly;
+                slider(e->scrolly, slines, 0, color, NULL);
+                if(pos != e->scrolly) e->cy = e->scrolly; 
+            }
+            if(wasvertical) poplist();
+        }
+        
+        return result;
+    }
 
     void rect_(float x, float y, float w, float h, int usetc = -1) 
     {
-        static const GLint tc[4][2] = {{0, 0}, {1, 0}, {1, 1}, {0, 1}};
-        if(usetc>=0) glTexCoord2iv(tc[usetc]); 
+        static const GLfloat tc[4][2] = {{0, 0}, {1, 0}, {1, 1}, {0, 1}};
+        if(usetc>=0) glTexCoord2fv(tc[usetc]); 
         glVertex2f(x, y);
-        if(usetc>=0) glTexCoord2iv(tc[(usetc+1)%4]);
+        if(usetc>=0) glTexCoord2fv(tc[(usetc+1)%4]);
         glVertex2f(x + w, y);
-        if(usetc>=0) glTexCoord2iv(tc[(usetc+2)%4]);
+        if(usetc>=0) glTexCoord2fv(tc[(usetc+2)%4]);
         glVertex2f(x + w, y + h);
-        if(usetc>=0) glTexCoord2iv(tc[(usetc+3)%4]);
+        if(usetc>=0) glTexCoord2fv(tc[(usetc+3)%4]);
         glVertex2f(x, y + h);
         xtraverts += 4;
     }
@@ -464,7 +426,7 @@ struct gui : g3d_gui
         defaultshader->set();
     }
 
-    void icon_(Texture *t, bool overlaid, bool tiled, int x, int y, int size, bool hit) 
+    void icon_(Texture *t, bool overlaid, bool tiled, int x, int y, int size, bool hit, int rotate = 0, int xoff = 0, int yoff = 0, Texture *glowtex = NULL, const vec &glowcolor = vec(1, 1, 1)) 
     {
         float xs, ys, xt, yt;
         if(tiled)
@@ -501,22 +463,43 @@ struct gui : g3d_gui
             if(!rgbonlyshader) rgbonlyshader = lookupshaderbyname("rgbonly");
             rgbonlyshader->set();
         }
-        if(hit) glColor3f(1, 0.5f, 0.5f);
-        else if(overlaid) glColor3f(1, 1, 1);
-        else glColor3fv(light.v);
-        glBindTexture(GL_TEXTURE_2D, t->gl);
+        float tc[4][2] = { { 0, 0 }, { 1, 0 }, { 1, 1 }, { 0, 1 } };
+        if(rotate)
+        {
+            if((rotate&5) == 1) { swap(xoff, yoff); loopk(4) swap(tc[k][0], tc[k][1]); }
+            if(rotate >= 2 && rotate <= 4) { xoff *= -1; loopk(4) tc[k][0] *= -1; }
+            if(rotate <= 2 || rotate == 5) { yoff *= -1; loopk(4) tc[k][1] *= -1; }
+        }
+        loopk(4) { tc[k][0] = tc[k][0]/xt - float(xoff)/t->xs; tc[k][1] = tc[k][1]/yt - float(yoff)/t->ys; }
+        vec color = hit ? vec(1, 0.5f, 0.5f) : (overlaid ? vec(1, 1, 1) : light);
+        glColor3fv(color.v);
+        glBindTexture(GL_TEXTURE_2D, t->id);
         glBegin(GL_QUADS);
-        glTexCoord2f(0.0f,    0.0f);    glVertex2f(x,    y);
-        glTexCoord2f(1.0f/xt, 0.0f);    glVertex2f(x+xs, y);
-        glTexCoord2f(1.0f/xt, 1.0f/yt); glVertex2f(x+xs, y+ys);
-        glTexCoord2f(0.0f,    1.0f/yt); glVertex2f(x,    y+ys);
+        glTexCoord2fv(tc[0]); glVertex2f(x,    y);
+        glTexCoord2fv(tc[1]); glVertex2f(x+xs, y);
+        glTexCoord2fv(tc[2]); glVertex2f(x+xs, y+ys);
+        glTexCoord2fv(tc[3]); glVertex2f(x,    y+ys);
         glEnd();
+        if(glowtex)
+        {
+            if(hit || overlaid) { loopk(3) color[k] *= glowcolor[k]; glColor3fv(color.v); }
+            else glColor3fv(glowcolor.v);
+            glBlendFunc(GL_SRC_ALPHA, GL_ONE);
+            glBindTexture(GL_TEXTURE_2D, glowtex->id);
+            glBegin(GL_QUADS);
+            glTexCoord2fv(tc[0]); glVertex2f(x,    y);
+            glTexCoord2fv(tc[1]); glVertex2f(x+xs, y);
+            glTexCoord2fv(tc[2]); glVertex2f(x+xs, y+ys);
+            glTexCoord2fv(tc[3]); glVertex2f(x,    y+ys);
+            glEnd();
+            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
+        }
         if(tiled) defaultshader->set();
         if(overlaid) 
         {
             if(!overlaytex) overlaytex = textureload("data/guioverlay.png");
             glColor3fv(light.v);
-            glBindTexture(GL_TEXTURE_2D, overlaytex->gl);
+            glBindTexture(GL_TEXTURE_2D, overlaytex->id);
             glBegin(GL_QUADS);
             rect_(x, y, xs, ys, 0);
             glEnd();
@@ -529,7 +512,7 @@ struct gui : g3d_gui
         {
             if(!slidertex) slidertex = textureload("data/guislider.png");
             glEnable(GL_TEXTURE_2D);
-            glBindTexture(GL_TEXTURE_2D, slidertex->gl);
+            glBindTexture(GL_TEXTURE_2D, slidertex->id);
             glBegin(GL_QUADS);
             if(percent < 0.99f) 
             {
@@ -583,7 +566,7 @@ struct gui : g3d_gui
     void skin_(int x, int y, int gapw, int gaph, int start, int n)//int vleft, int vright, int vtop, int vbottom, int start, int n) 
     {
         if(!skintex) skintex = textureload("data/guiskin.png");
-        glBindTexture(GL_TEXTURE_2D, skintex->gl);
+        glBindTexture(GL_TEXTURE_2D, skintex->id);
         int gapx1 = INT_MAX, gapy1 = INT_MAX, gapx2 = INT_MAX, gapy2 = INT_MAX;
         float wscale = 1.0f/(SKIN_W*SKIN_SCALE), hscale = 1.0f/(SKIN_H*SKIN_SCALE);
         
@@ -641,10 +624,10 @@ struct gui : g3d_gui
                             tright = tleft+xstep*wscale;
                         }
                         if(!quads) { quads = true; glBegin(GL_QUADS); }
-                        glTexCoord2f(tleft,  ttop);    glVertex2i(xo,       yo);
-                        glTexCoord2f(tright, ttop);    glVertex2i(xo+xstep, yo);
-                        glTexCoord2f(tright, tbottom); glVertex2i(xo+xstep, yo+ystep);
-                        glTexCoord2f(tleft,  tbottom); glVertex2i(xo,       yo+ystep);
+                        glTexCoord2f(tleft,  ttop);    glVertex2f(xo,       yo);
+                        glTexCoord2f(tright, ttop);    glVertex2f(xo+xstep, yo);
+                        glTexCoord2f(tright, tbottom); glVertex2f(xo+xstep, yo+ystep);
+                        glTexCoord2f(tleft,  tbottom); glVertex2f(xo,       yo+ystep);
                         xtraverts += 4;
                         if(!(p.flags&0x01)) break;
                         xo += xstep;
@@ -660,34 +643,38 @@ struct gui : g3d_gui
         if(!gui2d) glDepthFunc(GL_ALWAYS);
     } 
 
-    vec origin, scale;
+    vec origin, scale, *savedorigin;
     float dist;
     g3d_callback *cb;
+    bool gui2d;
 
-    static float basescale;
+    static float basescale, maxscale;
     static bool passthrough;
     static vec light;
 
     void adjustscale()
     {
-        float aspect = float(screen->h)/float(screen->w), fit = 1.0f;
         int w = xsize + (skinx[2]-skinx[1])*SKIN_SCALE + (skinx[10]-skinx[9])*SKIN_SCALE, h = ysize + (skiny[8]-skiny[6])*SKIN_SCALE;
         if(tcurrent) h += ((skiny[5]-skiny[1])-(skiny[3]-skiny[2]))*SKIN_SCALE + FONTH-2*INSERT;
         else h += (skiny[5]-skiny[3])*SKIN_SCALE;
 
-        if(w*aspect*basescale>1) fit = 1.0f/(w*aspect*basescale);
-        if(h*basescale*fit>1) fit *= 1.0f/(h*basescale*fit);
+        float aspect = float(screen->h)/float(screen->w), fit = 1.0f;
+        if(w*aspect*basescale>1.0f) fit = 1.0f/(w*aspect*basescale);
+        if(h*basescale*fit>maxscale) fit *= maxscale/(h*basescale*fit);
         origin = vec(0.5f-((w-xsize)/2 - (skinx[2]-skinx[1])*SKIN_SCALE)*aspect*scale.x*fit, 0.5f + (0.5f*h-(skiny[8]-skiny[6])*SKIN_SCALE)*scale.y*fit, 0);
         scale = vec(aspect*scale.x*fit, scale.y*fit, 1);
     }
 
     void start(int starttime, float initscale, int *tab, bool allowinput)
     {	
-        if(gui2d) initscale *= 0.025f; 
+        if(gui2d) 
+        {
+            initscale *= 0.025f; 
+            if(allowinput) hascursor = true;
+        }
         basescale = initscale;
         if(layoutpass) scale.x = scale.y = scale.z = basescale*min((totalmillis-starttime)/300.0f, 1.0f);
         passthrough = scale.x<basescale || !allowinput;
-        if(allowinput) hascursor = true;
         curdepth = -1;
         curlist = -1;
         tpos = 0;
@@ -721,7 +708,7 @@ struct gui : g3d_gui
                 vec dir;
                 lightreaching(origin, light, dir, 0, 0.5f); 
                 float intensity = vec(yaw, 0.0f).dot(dir);
-                light.mul(1.0f + max(intensity, 0));
+                light.mul(1.0f + max(intensity, 0.0f));
             }
 
             skin_(curx-skinx[2]*SKIN_SCALE, cury-skiny[5]*SKIN_SCALE, xsize, ysize, 0, 9);
@@ -754,6 +741,7 @@ struct gui : g3d_gui
                     hitx = xaxis.dot(intersectionvec)/scale.x;
                     hity = -intersectionvec.z/scale.y;
                 }
+                if((mousebuttons & G3D_PRESSED) && (fabs(hitx-firstx) > 2 || fabs(hity - firsty) > 2)) mousebuttons |= G3D_DRAGGED;
                 if(intersects>=INTERSECT_MIDDLE && hitx>=-xsize/2 && hitx<=xsize/2 && hity<=0)
                 {
                     if(hity>=-ysize || (tcurrent && hity>=-ysize-(FONTH-2*INSERT)-((skiny[5]-skiny[1])-(skiny[3]-skiny[2]))*SKIN_SCALE && hitx<=tx-xsize/2))
@@ -815,20 +803,134 @@ const gui::patch gui::patches[] =
 };
 
 vector<gui::list> gui::lists;
-float gui::basescale, gui::hitx, gui::hity;
-bool gui::passthrough;
+float gui::basescale, gui::maxscale = 1, gui::hitx, gui::hity;
+bool gui::passthrough, gui::shouldmergehits = false, gui::shouldautotab = true;
 vec gui::light;
 int gui::curdepth, gui::curlist, gui::xsize, gui::ysize, gui::curx, gui::cury;
 int gui::ty, gui::tx, gui::tpos, *gui::tcurrent, gui::tcolor;
-static vector<gui> guis;
+static vector<gui> guis2d, guis3d;
 
-void g3d_addgui(g3d_callback *cb, vec &origin, bool follow)
+VARP(guipushdist, 1, 4, 64);
+
+bool menukey(int code, bool isdown, int cooked)
 {
-    if(follow) origin.z = player->o.z-(player->eyeheight-1);
-    gui &g = guis.add();
+    if(code==-1 && g3d_windowhit(isdown, true)) return true;
+    else if(code==-3 && g3d_windowhit(isdown, false)) return true;
+
+    editor *e = currentfocus();
+    if((fieldmode == FIELDSHOW) || !e)
+    {
+        if(windowhit) switch(code)
+        {
+            case -4: // window "management" 
+                if(isdown)
+                {
+                    if(windowhit->gui2d) 
+                    {
+                        vec origin = *guis2d.last().savedorigin;
+                        int i = windowhit - &guis2d[0];
+                        for(int j = guis2d.length()-1; j > i; j--) *guis2d[j].savedorigin = *guis2d[j-1].savedorigin;
+                        *windowhit->savedorigin = origin;
+                        if(guis2d.length() > 1)
+                        {
+                            if(camera1->o.dist(*windowhit->savedorigin) <= camera1->o.dist(*guis2d.last().savedorigin))
+                                windowhit->savedorigin->add(camdir);
+                        }
+                    }
+                    else windowhit->savedorigin->add(vec(camdir).mul(guipushdist));
+                }
+                return true;
+            case -5:
+                if(isdown)
+                {
+                    if(windowhit->gui2d)
+                    {
+                        vec origin = *guis2d[0].savedorigin;
+                        loopj(guis2d.length()-1) *guis2d[j].savedorigin = *guis2d[j + 1].savedorigin;
+                        *guis2d.last().savedorigin = origin;
+                        if(guis2d.length() > 1)
+                        {
+                            if(camera1->o.dist(*guis2d.last().savedorigin) >= camera1->o.dist(*guis2d[0].savedorigin))
+                                guis2d.last().savedorigin->sub(camdir);
+                        }
+                    }
+                    else windowhit->savedorigin->sub(vec(camdir).mul(guipushdist));
+                }
+                return true;
+        }
+
+        return false;
+    }
+    switch(code)
+    {
+        case SDLK_ESCAPE: //cancel editing without commit
+            if(isdown) fieldmode = FIELDABORT;
+            return true;
+        case SDLK_RETURN:
+        case SDLK_TAB:
+            if(cooked && (e->maxy != 1)) break;
+        case SDLK_KP_ENTER:
+            if(isdown) fieldmode = FIELDCOMMIT; //signal field commit (handled when drawing field)
+            return true;
+        case SDLK_HOME:
+        case SDLK_END:
+        case SDLK_PAGEUP:
+        case SDLK_PAGEDOWN:
+        case SDLK_DELETE:
+        case SDLK_BACKSPACE:
+        case SDLK_UP:
+        case SDLK_DOWN:
+        case SDLK_LEFT:
+        case SDLK_RIGHT:
+        case -4:
+        case -5:
+            break;
+        default:
+            if(!cooked || (code<32)) return false;
+    }
+    if(!isdown) return true;
+    e->key(code, cooked);
+    return true;
+}
+
+void g3d_cursorpos(float &x, float &y)
+{
+    if(guis2d.length()) { x = cursorx; y = cursory; }
+    else x = y = 0.5f;
+}
+
+void g3d_resetcursor()
+{
+    cursorx = cursory = 0.5f;
+}
+
+bool g3d_movecursor(int dx, int dy)
+{
+    if(!guis2d.length() || !hascursor) return false;
+    const float CURSORSCALE = 500.0f;
+    cursorx = max(0.0f, min(1.0f, cursorx+dx*(screen->h/(screen->w*CURSORSCALE))));
+    cursory = max(0.0f, min(1.0f, cursory+dy/CURSORSCALE));
+    return true;
+}
+
+VARNP(guifollow, useguifollow, 0, 1, 1);
+VARNP(gui2d, usegui2d, 0, 1, 1);
+
+void g3d_addgui(g3d_callback *cb, vec &origin, int flags)
+{
+    bool gui2d = flags&GUI_FORCE_2D || (flags&GUI_2D && usegui2d);
+    if(!gui2d && flags&GUI_FOLLOW && useguifollow) origin.z = player->o.z-(player->eyeheight-1);
+    gui &g = (gui2d ? guis2d : guis3d).add();
     g.cb = cb;
     g.origin = origin;
+    g.savedorigin = &origin;
     g.dist = camera1->o.dist(g.origin);
+    g.gui2d = gui2d;
+}
+
+void g3d_limitscale(float scale)
+{
+    gui::maxscale = scale;
 }
 
 int g3d_sort(gui *a, gui *b) { return (int)(a->dist>b->dist)*2-1; }
@@ -836,86 +938,97 @@ int g3d_sort(gui *a, gui *b) { return (int)(a->dist>b->dist)*2-1; }
 bool g3d_windowhit(bool on, bool act)
 {
     extern int cleargui(int n);
-    if(act) mousebuttons |= (actionon=on) ? G3D_DOWN : G3D_UP;
-    else if(!on && windowhit) cleargui(1);
-    return gui2d ? hascursor : windowhit!=NULL;
-}
-
-char *g3d_fieldname()
-{
-    return fieldname;
+    if(act) 
+    {
+        if(on) { firstx = gui::hitx; firsty = gui::hity; }
+        mousebuttons |= (actionon=on) ? G3D_DOWN : G3D_UP;
+    } else if(!on && windowhit) cleargui(1);
+    return (guis2d.length() && hascursor) || (windowhit && !windowhit->gui2d);
 }
 
 void g3d_render()   
 {
-    windowhit = NULL;
+    windowhit = NULL;    
     if(actionon) mousebuttons |= G3D_PRESSED;
-    gui::reset();
-    guis.setsize(0);
     
+    gui::reset();
+    guis2d.setsize(0);
+    guis3d.setsize(0);
+ 
     // call all places in the engine that may want to render a gui from here, they call g3d_addgui()
     extern void g3d_texturemenu();
     
     g3d_texturemenu();
     g3d_mainmenu();
     cl->g3d_gamemenus();
+
+    guis2d.sort(g3d_sort);
+    guis3d.sort(g3d_sort);
     
-    guis.sort(g3d_sort);
-    
-    bool fieldfocus = (fieldpos>=0);
-    fieldactive = false;
+    readyeditors();
+    bool wasfocused = (fieldmode!=FIELDSHOW);
+    fieldsactive = false;
+
     hascursor = false;
 
-    if(guis.length())
+    layoutpass = true;
+    loopv(guis2d) guis2d[i].cb->gui(guis2d[i], true);
+    loopv(guis3d) guis3d[i].cb->gui(guis3d[i], true);
+    layoutpass = false;
+
+    if(guis2d.length() || guis3d.length())
     {
-        if(gui2d)
-        {
-            glMatrixMode(GL_PROJECTION);
-            glPushMatrix();
-            glLoadIdentity();
-            glOrtho(0, 1, 1, 0, -1, 1);
-
-            glMatrixMode(GL_MODELVIEW);
-            glPushMatrix();
-            glLoadIdentity();
-
-            glDisable(GL_DEPTH_TEST);
-        }
-        else
-        {
-            glDepthFunc(GL_ALWAYS);
-            glDepthMask(GL_FALSE);
-        }
         glEnable(GL_BLEND);
         glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
     }
 
-    layoutpass = true;
-    loopv(guis) guis[i].cb->gui(guis[i], true);
-    layoutpass = false;
-    loopvrev(guis) guis[i].cb->gui(guis[i], false);
-
-    if(guis.length())
+    if(guis3d.length())
     {
-        glDisable(GL_BLEND);
-        if(gui2d)
-        {
-            glEnable(GL_DEPTH_TEST);
-    
-            glMatrixMode(GL_PROJECTION);
-            glPopMatrix();
-            glMatrixMode(GL_MODELVIEW);
-            glPopMatrix();
-        }
-        else
-        { 
-            glDepthFunc(GL_LESS);
-            glDepthMask(GL_TRUE);
-        }
+        glDepthFunc(GL_ALWAYS);
+        glDepthMask(GL_FALSE);
+
+        loopvrev(guis3d) guis3d[i].cb->gui(guis3d[i], false);
+
+        glDepthFunc(GL_LESS);
+        glDepthMask(GL_TRUE);
     }
 
-    if(!fieldactive) fieldpos = -1; //no hit fields, so loose focus - mainly for menu closed
-    if((fieldpos>=0) != fieldfocus) SDL_EnableUNICODE(fieldpos>=0);
+    if(guis2d.length())
+    {
+        glMatrixMode(GL_PROJECTION);
+        glPushMatrix();
+        glLoadIdentity();
+        glOrtho(0, 1, 1, 0, -1, 1);
+
+        glMatrixMode(GL_MODELVIEW);
+        glPushMatrix();
+        glLoadIdentity();
+
+        glDisable(GL_DEPTH_TEST);
+
+        loopvrev(guis2d) guis2d[i].cb->gui(guis2d[i], false);
+
+        glEnable(GL_DEPTH_TEST);
+
+        glMatrixMode(GL_PROJECTION);
+        glPopMatrix();
+        glMatrixMode(GL_MODELVIEW);
+        glPopMatrix();
+    }
+
+    if(guis2d.length() || guis3d.length())
+    {
+        glDisable(GL_BLEND);
+    }
+
+    flusheditors();
+    if(!fieldsactive) fieldmode = FIELDSHOW; //didn't draw any fields, so loose focus - mainly for menu closed
+    if((fieldmode!=FIELDSHOW) != wasfocused) 
+    {
+        SDL_EnableUNICODE(fieldmode!=FIELDSHOW);
+        keyrepeat(fieldmode!=FIELDSHOW || editmode);
+    }
     
     mousebuttons = 0;
 }
+
