@@ -4,7 +4,6 @@
 #include "pch.h"
 #include "engine.h"
 
-void itoa(char *s, int i) { s_sprintf(s)("%d", i); }
 char *exchangestr(char *o, const char *n) { delete[] o; return newstring(n); }
 
 typedef hashtable<const char *, ident> identtable;
@@ -15,7 +14,7 @@ bool overrideidents = false, persistidents = true;
 
 void clearstack(ident &id)
 {
-    identstack *stack = id._stack;
+    identstack *stack = id.stack;
     while(stack)
     {
         delete[] stack->action;
@@ -23,64 +22,77 @@ void clearstack(ident &id)
         stack = stack->next;
         delete tmp;
     }
-    id._stack = NULL;
+    id.stack = NULL;
 }
 
 void clear_command()
 {
-    enumerate(*idents, ident, i, if(i._type==ID_ALIAS) { DELETEA(i._name); DELETEA(i._action); if(i._stack) clearstack(i); });
+    enumerate(*idents, ident, i, if(i.type==ID_ALIAS) { DELETEA(i.name); DELETEA(i.action); if(i.stack) clearstack(i); });
     if(idents) idents->clear();
 }
 
 void clearoverrides()
 {
     enumerate(*idents, ident, i,
-        if(i._override!=NO_OVERRIDE)
+        if(i.override!=NO_OVERRIDE)
         {
-            switch(i._type)
+            switch(i.type)
             {
                 case ID_ALIAS: 
-                    if(i._action[0]) i._action = exchangestr(i._action, ""); 
+                    if(i.action[0]) i.action = exchangestr(i.action, ""); 
                     break;
                 case ID_VAR: 
-                    *i._storage = i._override;
+                    *i.storage.i = i.overrideval.i;
+                    i.changed();
+                    break;
+                case ID_FVAR:
+                    *i.storage.f = i.overrideval.f;
+                    i.changed();
+                    break;
+                case ID_SVAR:
+                    delete[] *i.storage.s;
+                    *i.storage.s = i.overrideval.s;
                     i.changed();
                     break;
             }
-            i._override = NO_OVERRIDE;
+            i.override = NO_OVERRIDE;
         });
 }
 
 void pushident(ident &id, char *val)
 {
-    if(id._type != ID_ALIAS) return;
+    if(id.type != ID_ALIAS) return;
     identstack *stack = new identstack;
-    stack->action = id._isexecuting==id._action ? newstring(id._action) : id._action;
-    stack->next = id._stack;
-    id._stack = stack;
-    id._action = val;
+    stack->action = id.isexecuting==id.action ? newstring(id.action) : id.action;
+    stack->next = id.stack;
+    id.stack = stack;
+    id.action = val;
 }
 
 void popident(ident &id)
 {
-    if(id._type != ID_ALIAS || !id._stack) return;
-    if(id._action != id._isexecuting) delete[] id._action;
-    identstack *stack = id._stack;
-    id._action = stack->action;
-    id._stack = stack->next;
+    if(id.type != ID_ALIAS || !id.stack) return;
+    if(id.action != id.isexecuting) delete[] id.action;
+    identstack *stack = id.stack;
+    id.action = stack->action;
+    id.stack = stack->next;
     delete stack;
 }
 
-void pusha(char *name, char *action)
+ident *newident(const char *name)
 {
     ident *id = idents->access(name);
     if(!id)
     {
-        name = newstring(name);
-        ident init(ID_ALIAS, name, newstring(""), persistidents);
-        id = idents->access(name, &init);
+        ident init(ID_ALIAS, newstring(name), newstring(""), persistidents ? IDF_PERSIST : 0);
+        id = idents->access(init.name, &init);
     }
-    pushident(*id, action);
+    return id;
+}
+
+void pusha(const char *name, char *action)
+{
+    pushident(*newident(name), action);
 }
 
 void push(char *name, char *action)
@@ -102,24 +114,28 @@ void aliasa(const char *name, char *action)
     ident *b = idents->access(name);
     if(!b) 
     {
-        ident b(ID_ALIAS, newstring(name), action, persistidents);
-        if(overrideidents) b._override = OVERRIDDEN;
-        idents->access(b._name, &b);
+        ident b(ID_ALIAS, newstring(name), action, persistidents ? IDF_PERSIST : 0);
+        if(overrideidents) b.override = OVERRIDDEN;
+        idents->access(b.name, &b);
     }
-    else if(b->_type != ID_ALIAS)
+    else if(b->type != ID_ALIAS)
     {
-        conoutf("cannot redefine builtin %s with an alias", name);
+        conoutf(CON_ERROR, "cannot redefine builtin %s with an alias", name);
         delete[] action;
     }
     else 
     {
-        if(b->_action != b->_isexecuting) delete[] b->_action;
-        b->_action = action;
-        if(overrideidents) b->_override = OVERRIDDEN;
+        if(b->action != b->isexecuting) delete[] b->action;
+        b->action = action;
+        if(overrideidents) b->override = OVERRIDDEN;
         else 
         {
-            if(b->_override != NO_OVERRIDE) b->_override = NO_OVERRIDE;
-            if(b->_persist != persistidents) b->_persist = persistidents;
+            if(b->override != NO_OVERRIDE) b->override = NO_OVERRIDE;
+            if(persistidents)
+            {
+                if(!(b->flags & IDF_PERSIST)) b->flags |= IDF_PERSIST;
+            }
+            else if(b->flags & IDF_PERSIST) b->flags &= ~IDF_PERSIST;
         }
     }
 }
@@ -130,45 +146,87 @@ COMMAND(alias, "ss");
 
 // variable's and commands are registered through globals, see cube.h
 
-int variable(const char *name, int min, int cur, int max, int *storage, void (*fun)(), bool persist)
+int variable(const char *name, int min, int cur, int max, int *storage, void (*fun)(), int flags)
 {
     if(!idents) idents = new identtable;
-    ident v(ID_VAR, name, min, cur, max, storage, (void *)fun, persist);
+    ident v(ID_VAR, name, min, cur, max, storage, (void *)fun, flags);
     idents->access(name, &v);
     return cur;
 }
 
-#define GETVAR(id, name, retval) \
+float fvariable(const char *name, float cur, float *storage, void (*fun)(), int flags)
+{
+    if(!idents) idents = new identtable;
+    ident v(ID_FVAR, name, cur, storage, (void *)fun, flags);
+    idents->access(name, &v);
+    return cur;
+}
+
+char *svariable(const char *name, const char *cur, char **storage, void (*fun)(), int flags)
+{
+    if(!idents) idents = new identtable;
+    ident v(ID_SVAR, name, newstring(cur), storage, (void *)fun, flags);
+    idents->access(name, &v);
+    return v.val.s;
+}
+
+#define _GETVAR(id, vartype, name, retval) \
     ident *id = idents->access(name); \
-    if(!id || id->_type!=ID_VAR) return retval;
+    if(!id || id->type!=vartype) return retval;
+#define GETVAR(id, name, retval) _GETVAR(id, ID_VAR, name, retval)
 void setvar(const char *name, int i, bool dofunc) 
 { 
     GETVAR(id, name, );
-    *id->_storage = i; 
+    *id->storage.i = clamp(i, id->minval, id->maxval); 
     if(dofunc) id->changed();
 } 
+void setfvar(const char *name, float f, bool dofunc)
+{
+    _GETVAR(id, ID_FVAR, name, );
+    *id->storage.f = f;
+    if(dofunc) id->changed();
+}
+void setsvar(const char *name, const char *str, bool dofunc)
+{
+    _GETVAR(id, ID_SVAR, name, );
+    *id->storage.s = exchangestr(*id->storage.s, str);
+    if(dofunc) id->changed();
+}
 int getvar(const char *name) 
 { 
     GETVAR(id, name, 0);
-    return *id->_storage;
+    return *id->storage.i;
 }
 int getvarmin(const char *name) 
 { 
     GETVAR(id, name, 0);
-    return id->_min;
+    return id->minval;
 }
 int getvarmax(const char *name) 
 { 
     GETVAR(id, name, 0);
-    return id->_max;
+    return id->maxval;
 }
 bool identexists(const char *name) { return idents->access(name)!=NULL; }
 ident *getident(const char *name) { return idents->access(name); }
 
+void touchvar(const char *name)
+{
+    ident *id = idents->access(name);
+    if(id) switch(id->type)
+    {
+        case ID_VAR:
+        case ID_FVAR:
+        case ID_SVAR:
+            id->changed();
+            break;
+    }
+}
+
 const char *getalias(const char *name)
 {
     ident *i = idents->access(name);
-    return i && i->_type==ID_ALIAS ? i->_action : "";
+    return i && i->type==ID_ALIAS ? i->action : "";
 }
 
 bool addcommand(const char *name, void (*fun)(), const char *narg)
@@ -250,7 +308,7 @@ char *parseexp(const char *&p, int right)          // parse any nested set of ()
         else if(!c) 
         { 
             p--;
-            conoutf("missing \"%c\"", right);
+            conoutf(CON_ERROR, "missing \"%c\"", right);
             wordbuf.setsize(0); 
             bufnest--;
             return NULL; 
@@ -278,12 +336,14 @@ char *parseexp(const char *&p, int right)          // parse any nested set of ()
 char *lookup(char *n)                           // find value of ident referenced with $ in exp
 {
     ident *id = idents->access(n+1);
-    if(id) switch(id->_type)
+    if(id) switch(id->type)
     {
-        case ID_VAR: { string t; itoa(t, *(id->_storage)); return exchangestr(n, t); }
-        case ID_ALIAS: return exchangestr(n, id->_action);
+        case ID_VAR: { s_sprintfd(t)("%d", *id->storage.i); return exchangestr(n, t); }
+        case ID_FVAR: { s_sprintfd(t)("%f", *id->storage.f); return exchangestr(n, t); }
+        case ID_SVAR: return exchangestr(n, *id->storage.s);
+        case ID_ALIAS: return exchangestr(n, id->action);
     }
-    conoutf("unknown alias lookup: %s", n+1);
+    conoutf(CON_ERROR, "unknown alias lookup: %s", n+1);
     return n;
 }
 
@@ -347,7 +407,7 @@ char *executeret(const char *p)               // all evaluation happens here, re
     const int MAXWORDS = 25;                    // limit, remove
     char *w[MAXWORDS];
     char *retval = NULL;
-    #define setretval(v) { char *rv = v; if(rv) retval = rv; commandret = NULL; }
+    #define setretval(v) { char *rv = v; if(rv) retval = rv; }
     for(bool cont = true; cont;)                // for each ; seperated statement
     {
         int numargs = MAXWORDS;
@@ -363,7 +423,6 @@ char *executeret(const char *p)               // all evaluation happens here, re
         p += strcspn(p, ";\n\0");
         cont = *p++!=0;                         // more statements if this isn't the end of the string
         char *c = w[0];
-        if(*c=='/') c++;                        // strip irc-style command prefix
         if(!*c) continue;                       // empty statement
         
         DELETEA(retval);
@@ -378,16 +437,15 @@ char *executeret(const char *p)               // all evaluation happens here, re
             ident *id = idents->access(c);
             if(!id)
             {
-                if(!parseint(c) && *c!='0')
-                    conoutf("unknown command: %s", c);
+                if(!isdigit(*c) && ((*c!='+' && *c!='-' && *c!='.') || !isdigit(c[1]))) 
+                    conoutf(CON_ERROR, "unknown command: %s", c);
                 setretval(newstring(c));
             }
-            else switch(id->_type)
+            else switch(id->type)
             {
                 case ID_CCOMMAND:
                 case ID_COMMAND:                     // game defined commands
                 {   
-                    SAUERBRATEN_COMMAND_ENTRY(c, w[1], w[2], w[3]);
                     void *v[MAXWORDS];
                     union
                     {
@@ -396,96 +454,107 @@ char *executeret(const char *p)               // all evaluation happens here, re
                     } nstor[MAXWORDS];
                     int n = 0, wn = 0;
                     char *cargs = NULL;
-                    if(id->_type==ID_CCOMMAND) v[n++] = id->self;
-                    for(const char *a = id->_narg; *a; a++) switch(*a)
+                    if(id->type==ID_CCOMMAND) v[n++] = id->self;
+                    for(const char *a = id->narg; *a; a++) switch(*a)
                     {
                         case 's':                                 v[n] = w[++wn];     n++; break;
                         case 'i': nstor[n].i = parseint(w[++wn]); v[n] = &nstor[n].i; n++; break;
                         case 'f': nstor[n].f = atof(w[++wn]);     v[n] = &nstor[n].f; n++; break;
-                        case 'D': nstor[n].i = addreleaseaction(id->_name) ? 1 : 0; v[n] = &nstor[n].i; n++; break;
+                        case 'D': nstor[n].i = addreleaseaction(id->name) ? 1 : 0; v[n] = &nstor[n].i; n++; break;
                         case 'V': v[n++] = w+1; nstor[n].i = numargs-1; v[n] = &nstor[n].i; n++; break;
                         case 'C': if(!cargs) cargs = conc(w+1, numargs-1, true); v[n++] = cargs; break;
                         default: fatal("builtin declared with illegal type");
                     }
                     switch(n)
                     {
-                        case 0: ((void (__cdecl *)()                                      )id->_fun)();                             break;
-                        case 1: ((void (__cdecl *)(void *)                                )id->_fun)(v[0]);                         break;
-                        case 2: ((void (__cdecl *)(void *, void *)                        )id->_fun)(v[0], v[1]);                   break;
-                        case 3: ((void (__cdecl *)(void *, void *, void *)                )id->_fun)(v[0], v[1], v[2]);             break;
-                        case 4: ((void (__cdecl *)(void *, void *, void *, void *)        )id->_fun)(v[0], v[1], v[2], v[3]);       break;
-                        case 5: ((void (__cdecl *)(void *, void *, void *, void *, void *))id->_fun)(v[0], v[1], v[2], v[3], v[4]); break;
-                        case 6: ((void (__cdecl *)(void *, void *, void *, void *, void *, void *))id->_fun)(v[0], v[1], v[2], v[3], v[4], v[5]); break;
-                        case 7: ((void (__cdecl *)(void *, void *, void *, void *, void *, void *, void *))id->_fun)(v[0], v[1], v[2], v[3], v[4], v[5], v[6]); break;
-                        case 8: ((void (__cdecl *)(void *, void *, void *, void *, void *, void *, void *, void *))id->_fun)(v[0], v[1], v[2], v[3], v[4], v[5], v[6], v[7]); break;
+                        case 0: ((void (__cdecl *)()                                      )id->fun)();                             break;
+                        case 1: ((void (__cdecl *)(void *)                                )id->fun)(v[0]);                         break;
+                        case 2: ((void (__cdecl *)(void *, void *)                        )id->fun)(v[0], v[1]);                   break;
+                        case 3: ((void (__cdecl *)(void *, void *, void *)                )id->fun)(v[0], v[1], v[2]);             break;
+                        case 4: ((void (__cdecl *)(void *, void *, void *, void *)        )id->fun)(v[0], v[1], v[2], v[3]);       break;
+                        case 5: ((void (__cdecl *)(void *, void *, void *, void *, void *))id->fun)(v[0], v[1], v[2], v[3], v[4]); break;
+                        case 6: ((void (__cdecl *)(void *, void *, void *, void *, void *, void *))id->fun)(v[0], v[1], v[2], v[3], v[4], v[5]); break;
+                        case 7: ((void (__cdecl *)(void *, void *, void *, void *, void *, void *, void *))id->fun)(v[0], v[1], v[2], v[3], v[4], v[5], v[6]); break;
+                        case 8: ((void (__cdecl *)(void *, void *, void *, void *, void *, void *, void *, void *))id->fun)(v[0], v[1], v[2], v[3], v[4], v[5], v[6], v[7]); break;
                         default: fatal("builtin declared with too many args (use V?)");
                     }
                     if(cargs) delete[] cargs;
                     setretval(commandret);
-                    SAUERBRATEN_COMMAND_RETURN(c);
+                    commandret = NULL;
                     break;
                 }
 
                 case ID_VAR:                        // game defined variables 
-                    SAUERBRATEN_VAR_ENTRY(c, w[1]);
-                    if(!w[1][0]) conoutf("%s = %d", c, *id->_storage);      // var with no value just prints its current value
-                    else if(id->_min>id->_max) conoutf("variable %s is read-only", id->_name);
+                    if(numargs <= 1) conoutf("%s = %d", c, *id->storage.i);      // var with no value just prints its current value
+                    else if(id->minval>id->maxval) conoutf(CON_ERROR, "variable %s is read-only", id->name);
                     else
                     {
-                        if(overrideidents)
-                        {
-                            if(id->_persist)
-                            {
-                                conoutf("cannot override persistent variable %s", id->_name);
-                                break;
-                            }
-                            if(id->_override==NO_OVERRIDE) id->_override = *id->_storage;
-                        }
-                        else if(id->_override!=NO_OVERRIDE) id->_override = NO_OVERRIDE;
+                        #define OVERRIDEVAR(saveval, resetval) \
+                            if(overrideidents || id->flags&IDF_OVERRIDE) \
+                            { \
+                                if(id->flags&IDF_PERSIST) \
+                                { \
+                                    conoutf(CON_ERROR, "cannot override persistent variable %s", id->name); \
+                                    break; \
+                                } \
+                                if(id->override==NO_OVERRIDE) { saveval; id->override = OVERRIDDEN; } \
+                            } \
+                            else if(id->override!=NO_OVERRIDE) { resetval; id->override = NO_OVERRIDE; }
+                        OVERRIDEVAR(id->overrideval.i = *id->storage.i, )
                         int i1 = parseint(w[1]);
-                        if(i1<id->_min || i1>id->_max)
+                        if(i1<id->minval || i1>id->maxval)
                         {
-                            i1 = i1<id->_min ? id->_min : id->_max;                // clamp to valid range
-                            conoutf("valid range for %s is %d..%d", id->_name, id->_min, id->_max);
+                            i1 = i1<id->minval ? id->minval : id->maxval;                // clamp to valid range
+                            conoutf(CON_ERROR, "valid range for %s is %d..%d", id->name, id->minval, id->maxval);
                         }
-                        *id->_storage = i1;
+                        *id->storage.i = i1;
                         id->changed();                                             // call trigger function if available
                     }
-                    SAUERBRATEN_VAR_RETURN(c, *id->_storage);
                     break;
-                    
+                  
+                case ID_FVAR:
+                    if(numargs <= 1) conoutf("%s = %f", c, *id->storage.f);
+                    else
+                    {
+                        OVERRIDEVAR(id->overrideval.f = *id->storage.f, );
+                        *id->storage.f = atof(w[1]);
+                        id->changed();
+                    }
+                    break;
+ 
+                case ID_SVAR:
+                    if(numargs <= 1) conoutf(strchr(*id->storage.s, '"') ? "%s = [%s]" : "%s = \"%s\"", c, *id->storage.s);
+                    else
+                    {
+                        OVERRIDEVAR(id->overrideval.s = *id->storage.s, delete[] id->overrideval.s);
+                        *id->storage.s = newstring(w[1]);
+                        id->changed();
+                    }
+                    break;
+                        
                 case ID_ALIAS:                              // alias, also used as functions and (global) variables
                 {
-                    SAUERBRATEN_ALIAS_ENTRY(c, w[1], w[2], w[3]);
                     static vector<ident *> argids;
                     for(int i = 1; i<numargs; i++)
                     {
                         if(i > argids.length())
                         {
                             s_sprintfd(argname)("arg%d", i);
-                            ident *id = idents->access(argname);
-                            if(!id)
-                            {
-                                ident init(ID_ALIAS, newstring(argname), newstring(""), persistidents);
-                                id = idents->access(init._name, &init);
-                            }
-                            argids.add(id);
+                            argids.add(newident(argname));
                         }
                         pushident(*argids[i-1], w[i]); // set any arguments as (global) arg values so functions can access them
-                        w[i] = NULL;
                     }
                     _numargs = numargs-1;
                     bool wasoverriding = overrideidents;
-                    if(id->_override!=NO_OVERRIDE) overrideidents = true;
-                    char *wasexecuting = id->_isexecuting;
-                    id->_isexecuting = id->_action;
-                    setretval(executeret(id->_action));
-                    if(id->_isexecuting != id->_action && id->_isexecuting != wasexecuting) delete[] id->_isexecuting;
-                    id->_isexecuting = wasexecuting;
+                    if(id->override!=NO_OVERRIDE) overrideidents = true;
+                    char *wasexecuting = id->isexecuting;
+                    id->isexecuting = id->action;
+                    setretval(executeret(id->action));
+                    if(id->isexecuting != id->action && id->isexecuting != wasexecuting) delete[] id->isexecuting;
+                    id->isexecuting = wasexecuting;
                     overrideidents = wasoverriding;
                     for(int i = 1; i<numargs; i++) popident(*argids[i-1]);
-                    SAUERBRATEN_ALIAS_RETURN(c);
-                    break;
+                    continue;
                 }
             }
         }
@@ -515,7 +584,7 @@ bool execfile(const char *cfgfile)
 
 void exec(const char *cfgfile)
 {
-    if(!execfile(cfgfile)) conoutf("could not read \"%s\"", cfgfile);
+    if(!execfile(cfgfile)) conoutf(CON_ERROR, "could not read \"%s\"", cfgfile);
 }
 
 void writecfg()
@@ -527,18 +596,20 @@ void writecfg()
     fprintf(f, "\n");
     writecrosshairs(f);
     enumerate(*idents, ident, id,
-        if(id._type==ID_VAR && id._persist)
+        if(id.flags&IDF_PERSIST) switch(id.type)
         {
-            fprintf(f, "%s %d\n", id._name, *id._storage);
+            case ID_VAR: fprintf(f, "%s %d\n", id.name, *id.storage.i); break;
+            case ID_FVAR: fprintf(f, "%s %f\n", id.name, *id.storage.f); break;
+            case ID_SVAR: fprintf(f, "%s [%s]\n", id.name, *id.storage.s); break;
         }
     );
     fprintf(f, "\n");
     writebinds(f);
     fprintf(f, "\n");
     enumerate(*idents, ident, id,
-        if(id._type==ID_ALIAS && id._persist && id._override==NO_OVERRIDE && !strstr(id._name, "nextmap_") && id._action[0])
+        if(id.type==ID_ALIAS && id.flags&IDF_PERSIST && id.override==NO_OVERRIDE && !strstr(id.name, "nextmap_") && id.action[0])
         {
-            fprintf(f, "\"%s\" = [%s]\n", id._name, id._action);
+            fprintf(f, "\"%s\" = [%s]\n", id.name, id.action);
         }
     );
     fprintf(f, "\n");
@@ -551,12 +622,22 @@ COMMAND(writecfg, "");
 // below the commands that implement a small imperative language. thanks to the semantics of
 // () and [] expressions, any control construct can be defined trivially.
 
-void intset(char *name, int v) { string b; itoa(b, v); alias(name, b); }
-void intret            (int v) { string b; itoa(b, v); commandret = newstring(b); }
+void intret(int v) { s_sprintfd(b)("%d", v); commandret = newstring(b); }
 
 ICOMMAND(if, "sss", (char *cond, char *t, char *f), commandret = executeret(cond[0]!='0' ? t : f));
-
-ICOMMAND(loop, "sis", (char *var, int *n, char *body), loopi(*n) { intset(var, i); execute(body); });
+ICOMMAND(loop, "sis", (char *var, int *n, char *body), 
+{
+    if(*n<=0) return;
+    ident *id = newident(var);
+    if(id->type!=ID_ALIAS) return;
+    loopi(*n)
+    {
+        if(i) sprintf(id->action, "%d", i);
+        else pushident(*id, newstring("0", 16));
+        execute(body); 
+    } 
+    popident(*id);
+});
 ICOMMAND(while, "ss", (char *cond, char *body), while(execute(cond)) execute(body));    // can't get any simpler than this :)
 
 void concat(const char *s) { commandret = newstring(s); }
@@ -594,18 +675,32 @@ void format(char **args, int *numargs)
 #define whitespaceskip s += strspn(s, "\n\t ")
 #define elementskip *s=='"' ? (++s, s += strcspn(s, "\"\n\0"), s += *s=='"') : s += strcspn(s, "\n\t \0")
 
-void explodelist(char *s, vector<char *> &elems)
+void explodelist(const char *s, vector<char *> &elems)
 {
     whitespaceskip;
     while(*s)
     {
-        char *elem = s;
+        const char *elem = s;
         elementskip;
         elems.add(*elem=='"' ? newstring(elem+1, s-elem-(s[-1]=='"' ? 2 : 1)) : newstring(elem, s-elem));
         whitespaceskip;
     }
 }
 
+char *indexlist(const char *s, int pos)
+{
+    whitespaceskip;
+    loopi(pos) elementskip, whitespaceskip;
+    const char *e = s;
+    elementskip;
+    if(*e=='"')
+    {
+        e++;
+        if(s[-1]=='"') --s;
+    }
+    return newstring(e, s-e);
+}
+
 void listlen(char *s)
 {
     int n = 0;
@@ -616,17 +711,13 @@ void listlen(char *s)
 
 void at(char *s, int *pos)
 {
-    whitespaceskip;
-    loopi(*pos) elementskip, whitespaceskip;
-    char *e = s;
-    elementskip;
-    if(*e=='"') 
-    {
-        e++;
-        if(s[-1]=='"') --s;
-    }
-    *s = '\0';
-    result(e);
+    commandret = indexlist(s, *pos);
+}
+
+void substr(char *s, int *start, int *count)
+{
+    int len = strlen(s), offset = clamp(*start, 0, len);
+    commandret = newstring(&s[offset], *count <= 0 ? len - offset : min(*count, len - offset));
 }
 
 void getalias_(char *s)
@@ -640,6 +731,7 @@ COMMAND(result, "s");
 COMMAND(concatword, "V");
 COMMAND(format, "V");
 COMMAND(at, "si");
+COMMAND(substr, "sii");
 COMMAND(listlen, "s");
 COMMANDN(getalias, getalias_, "s");
 
@@ -652,8 +744,8 @@ void equal(int *a, int *b) { intret((int)(*a == *b)); }  COMMANDN(=, equal, "ii"
 void nequal(int *a, int *b) { intret((int)(*a != *b)); } COMMANDN(!=, nequal, "ii");
 void lt   (int *a, int *b) { intret((int)(*a < *b)); }   COMMANDN(<, lt, "ii");
 void gt   (int *a, int *b) { intret((int)(*a > *b)); }   COMMANDN(>, gt, "ii");
-void lte   (int *a, int *b) { intret((int)(*a <= *b)); } COMMANDN(<=, lte, "ii");
-void gte   (int *a, int *b) { intret((int)(*a >= *b)); } COMMANDN(>=, gte, "ii");
+void lte  (int *a, int *b) { intret((int)(*a <= *b)); } COMMANDN(<=, lte, "ii");
+void gte  (int *a, int *b) { intret((int)(*a >= *b)); } COMMANDN(>=, gte, "ii");
 void xora (int *a, int *b) { intret(*a ^ *b); }          COMMANDN(^, xora, "ii");
 void nota (int *a)         { intret(*a == 0); }          COMMANDN(!, nota, "i");
 void mina (int *a, int *b) { intret(min(*a, *b)); }      COMMANDN(min, mina, "ii");
@@ -669,10 +761,35 @@ void rndn(int *a)          { intret(*a>0 ? rnd(*a) : 0); }  COMMANDN(rnd, rndn,
 
 void strcmpa(char *a, char *b) { intret(strcmp(a,b)==0); }  COMMANDN(strcmp, strcmpa, "ss");
 
-ICOMMAND(echo, "C", (char *s), conoutf("\f1%s", s));
+ICOMMAND(echo, "C", (char *s), conoutf(CON_ECHO, "\f1%s", s));
 
 void strstra(char *a, char *b) { char *s = strstr(a, b); intret(s ? s-a : -1); } COMMANDN(strstr, strstra, "ss");
 
+char *strreplace(const char *s, const char *oldval, const char *newval)
+{
+    vector<char> buf;
+
+    int oldlen = strlen(oldval);
+    for(;;)
+    {
+        const char *found = strstr(s, oldval);
+        if(found)
+        {
+            while(s < found) buf.add(*s++);
+            for(const char *n = newval; *n; n++) buf.add(*n);
+            s = found + oldlen;
+        }
+        else 
+        {
+            while(*s) buf.add(*s++);
+            buf.add('\0');
+            return newstring(buf.getbuf(), buf.length());
+        }
+    }
+}
+
+void strreplacea(char *s, char *o, char *n) { commandret = strreplace(s, o, n); } COMMANDN(strreplace, strreplacea, "sss");
+
 struct sleepcmd
 {
     int millis;
