@@ -1,11 +1,24 @@
 // GL_ARB_vertex_program, GL_ARB_fragment_program
-extern PFNGLGENPROGRAMSARBPROC            glGenPrograms_;
-extern PFNGLDELETEPROGRAMSARBPROC         glDeletePrograms_;
-extern PFNGLBINDPROGRAMARBPROC            glBindProgram_;
-extern PFNGLPROGRAMSTRINGARBPROC          glProgramString_;
-extern PFNGLGETPROGRAMIVARBPROC           glGetProgramiv_;
-extern PFNGLPROGRAMENVPARAMETER4FARBPROC  glProgramEnvParameter4f_;
-extern PFNGLPROGRAMENVPARAMETER4FVARBPROC glProgramEnvParameter4fv_;
+extern PFNGLGENPROGRAMSARBPROC              glGenPrograms_;
+extern PFNGLDELETEPROGRAMSARBPROC           glDeletePrograms_;
+extern PFNGLBINDPROGRAMARBPROC              glBindProgram_;
+extern PFNGLPROGRAMSTRINGARBPROC            glProgramString_;
+extern PFNGLGETPROGRAMIVARBPROC             glGetProgramiv_;
+extern PFNGLPROGRAMENVPARAMETER4FARBPROC    glProgramEnvParameter4f_;
+extern PFNGLPROGRAMENVPARAMETER4FVARBPROC   glProgramEnvParameter4fv_;
+extern PFNGLENABLEVERTEXATTRIBARRAYARBPROC  glEnableVertexAttribArray_;
+extern PFNGLDISABLEVERTEXATTRIBARRAYARBPROC glDisableVertexAttribArray_;
+extern PFNGLVERTEXATTRIBPOINTERARBPROC      glVertexAttribPointer_;
+
+// GL_EXT_gpu_program_parameters
+#ifndef GL_EXT_gpu_program_parameters
+#define GL_EXT_gpu_program_parameters 1
+typedef void (APIENTRYP PFNGLPROGRAMENVPARAMETERS4FVEXTPROC) (GLenum target, GLuint index, GLsizei count, const GLfloat *params);
+typedef void (APIENTRYP PFNGLPROGRAMLOCALPARAMETERS4FVEXTPROC) (GLenum target, GLuint index, GLsizei count, const GLfloat *params);
+#endif
+
+extern PFNGLPROGRAMENVPARAMETERS4FVEXTPROC   glProgramEnvParameters4fv_;
+extern PFNGLPROGRAMLOCALPARAMETERS4FVEXTPROC glProgramLocalParameters4fv_;
 
 // GL_ARB_shading_language_100, GL_ARB_shader_objects, GL_ARB_fragment_shader, GL_ARB_vertex_shader
 extern PFNGLCREATEPROGRAMOBJECTARBPROC  glCreateProgramObject_;
@@ -54,12 +67,20 @@ struct LocalShaderParamState : ShaderParam
 
 struct ShaderParamState
 {
+    enum
+    {
+        CLEAN = 0,
+        INVALID,
+        DIRTY
+    };
+    
     const char *name;
     float val[4];
-    bool dirty, local;
+    bool local;
+    int dirty;
 
     ShaderParamState()
-        : name(NULL), dirty(false), local(false)
+        : name(NULL), local(false), dirty(INVALID)
     {
         memset(val, 0, sizeof(val));
     }
@@ -67,6 +88,8 @@ struct ShaderParamState
 
 enum 
 { 
+    SHADER_INVALID    = -1,
+
     SHADER_DEFAULT    = 0, 
     SHADER_NORMALSLMS = 1<<0, 
     SHADER_ENVMAP     = 1<<1,
@@ -74,7 +97,7 @@ enum
 };
 
 #define MAXSHADERDETAIL 3
-#define MAXVARIANTROWS 4
+#define MAXVARIANTROWS 5
 
 extern int shaderdetail;
 
@@ -84,22 +107,31 @@ struct Shader
 {
     static Shader *lastshader;
 
-    char *name;
+    char *name, *vsstr, *psstr;
     int type;
     GLuint vs, ps;
     GLhandleARB program, vsobj, psobj;
-    vector<LocalShaderParamState> defaultparams, extparams;
-    Shader *altshader, *fastshader[MAXSHADERDETAIL];
+    vector<LocalShaderParamState> defaultparams;
+    Shader *variantshader, *altshader, *fastshader[MAXSHADERDETAIL];
     vector<Shader *> variants[MAXVARIANTROWS];
-    LocalShaderParamState *extvertparams[RESERVEDSHADERPARAMS], *extpixparams[RESERVEDSHADERPARAMS];
-    bool used, native;
+    bool standard, used, native;
+    Shader *reusevs, *reuseps;
+    int numextparams;
+    LocalShaderParamState *extparams;
+    uchar *extvertparams, *extpixparams;
 
-    Shader() : name(NULL), type(SHADER_DEFAULT), vs(0), ps(0), program(0), vsobj(0), psobj(0), altshader(NULL), used(false), native(true)
+
+    Shader() : name(NULL), vsstr(NULL), psstr(NULL), type(SHADER_DEFAULT), vs(0), ps(0), program(0), vsobj(0), psobj(0), variantshader(NULL), altshader(NULL), standard(false), used(false), native(true), reusevs(NULL), reuseps(NULL), numextparams(0), extparams(NULL), extvertparams(NULL), extpixparams(NULL)
     {}
 
     ~Shader()
     {
         DELETEA(name);
+        DELETEA(vsstr);
+        DELETEA(psstr);
+        DELETEA(extparams);
+        DELETEA(extvertparams);
+        extpixparams = NULL;
     }
 
     void allocenvparams(Slot *slot = NULL);
@@ -107,6 +139,13 @@ struct Shader
     void setslotparams(Slot &slot);
     void bindprograms();
 
+    Shader *hasvariant(int col, int row = 0)
+    {
+        if(!this || renderpath==R_FIXEDFUNCTION) return NULL;
+        Shader *s = shaderdetail < MAXSHADERDETAIL ? fastshader[shaderdetail] : this;
+        return row>=0 && row<MAXVARIANTROWS && s->variants[row].inrange(col) ? s->variants[row][col] : NULL;
+    }
+
     Shader *variant(int col, int row = 0)
     {
         if(!this || renderpath==R_FIXEDFUNCTION) return this;
@@ -125,17 +164,36 @@ struct Shader
         lastshader->flushenvparams(slot);
         if(slot) lastshader->setslotparams(*slot);
     }
+
+    bool compile();
+    void cleanup(bool invalid = false);
 };
 
+#define SETSHADER(name) \
+    do { \
+        static Shader *name##shader = NULL; \
+        if(!name##shader) name##shader = lookupshaderbyname(#name); \
+        name##shader->set(); \
+    } while(0)
+
 // management of texture slots
 // each texture slot can have multiple texture frames, of which currently only the first is used
 // additional frames can be used for various shaders
 
 struct Texture
 {
+    enum
+    {
+        STUB,
+        TRANSIENT,
+        IMAGE,
+        CUBEMAP
+    };
+
     char *name;
-    int xs, ys, bpp;
-    GLuint gl;
+    int type, w, h, xs, ys, bpp, clamp;
+    bool mipmap, canreduce;
+    GLuint id;
     uchar *alphamask;
 
     Texture() : alphamask(NULL) {}
@@ -160,59 +218,82 @@ struct Slot
         int type;
         Texture *t;
         string name;
-        int rotation, xoffset, yoffset;
-        float scale;
         int combined;
     };
 
     vector<Tex> sts;
     Shader *shader;
     vector<ShaderParam> params;
-    bool loaded;
+    float scale;
+    int rotation, xoffset, yoffset;
+    float scrollS, scrollT;
+    vec glowcolor, pulseglowcolor;
+    float pulseglowspeed;
+    bool mtglowed, loaded;
+    uint texmask;
     char *autograss;
     Texture *grasstex, *thumbnail;
+
+    Slot() : autograss(NULL) { reset(); }
     
     void reset()
     {
         sts.setsize(0);
         shader = NULL;
         params.setsize(0);
+        scale = 1;
+        rotation = xoffset = yoffset = 0;
+        scrollS = scrollT = 0;
+        glowcolor = vec(1, 1, 1);
+        pulseglowcolor = vec(0, 0, 0);
+        pulseglowspeed = 0;
         loaded = false;
+        texmask = 0;
         DELETEA(autograss);
         grasstex = NULL;
         thumbnail = NULL;
     }
-    
-    Slot() : autograss(NULL) { reset(); }
+
+    void cleanup()
+    {
+        loaded = false;
+        grasstex = NULL;
+        thumbnail = NULL;
+        loopv(sts) 
+        {
+            Tex &t = sts[i];
+            t.t = NULL;
+            t.combined = -1;
+        }
+    }
 };
 
 struct cubemapside
 {
     GLenum target;
     const char *name;
+    bool flipx, flipy, swapxy;
 };
 
 extern cubemapside cubemapsides[6];
-
 extern Texture *notexture;
-
-extern Shader *defaultshader, *notextureshader, *nocolorshader, *foggedshader, *foggednotextureshader;
+extern Shader *defaultshader, *rectshader, *notextureshader, *nocolorshader, *foggedshader, *foggednotextureshader, *stdworldshader;
+extern int reservevpparams, maxvpenvparams, maxvplocalparams, maxfpenvparams, maxfplocalparams;
 
 extern Shader *lookupshaderbyname(const char *name);
-
 extern Texture *loadthumbnail(Slot &slot);
-
 extern void setslotshader(Slot &s);
-
 extern void setenvparamf(const char *name, int type, int index, float x = 0, float y = 0, float z = 0, float w = 0);
 extern void setenvparamfv(const char *name, int type, int index, const float *v);
-extern void flushenvparam(int type, int index, bool local = false);
+extern void flushenvparamf(const char *name, int type, int index, float x = 0, float y = 0, float z = 0, float w = 0);
+extern void flushenvparamfv(const char *name, int type, int index, const float *v);
 extern void setlocalparamf(const char *name, int type, int index, float x = 0, float y = 0, float z = 0, float w = 0);
 extern void setlocalparamfv(const char *name, int type, int index, const float *v);
-
+extern void invalidateenvparams(int type, int start, int count);
 extern ShaderParam *findshaderparam(Slot &s, const char *name, int type, int index);
 
 extern int maxtmus, nolights, nowater, nomasks;
+
 extern void inittmus();
 extern void resettmu(int n);
 extern void scaletmu(int n, int rgbscale, int alphascale = 0);
@@ -220,4 +301,11 @@ extern void colortmu(int n, float r = 0, float g = 0, float b = 0, float a = 0);
 extern void setuptmu(int n, const char *rgbfunc = NULL, const char *alphafunc = NULL);
 
 #define MAXDYNLIGHTS 5
+#define DYNLIGHTBITS 6
+#define DYNLIGHTMASK ((1<<DYNLIGHTBITS)-1)
+
+#define MAXBLURRADIUS 7
+
+extern void setupblurkernel(int radius, float sigma, float *weights, float *offsets);
+extern void setblurshader(int pass, int size, int radius, float *weights, float *offsets, GLenum target = GL_TEXTURE_2D);
 
